from typing import List

import numpy as np
import tensorflow as tf
from keras.models import Model, load_model
from keras.optimizers import Adam

from seg_tgce.data.oxford_pet.disturbance.model import (
    download_base_model,
    find_last_encoder_conv_layer,
    produce_disturbed_models,
)
from seg_tgce.data.oxford_pet.oxford_pet import get_data_multiple_annotators
from seg_tgce.loss.tgce import TcgeConfig, TcgeSs
from seg_tgce.metrics.dice_coefficient import DiceCoefficient
from seg_tgce.models.unet import unet_tgce
from seg_tgce.run.oxford_ma_runner.model_result import ModelResult
from seg_tgce.run.oxford_ma_runner.plotting import (
    epoch_progress_plotter,
    plot_losses_and_metrics,
)
from seg_tgce.run.runner import (
    Runner,
    RunningSessionParams,
    SessionPartialResults,
    SessionResults,
)


def fetch_models(noise_levels_snr: list[float]) -> List[Model]:
    model_path = download_base_model()
    model_ann = load_model(model_path, compile=False)

    last_conv_encoder_layer = find_last_encoder_conv_layer(model_ann)

    disturbance_models, measured_snr_values = produce_disturbed_models(
        noise_levels_snr, model_path, last_conv_encoder_layer
    )
    print(f"Measured snr values from produced models: {measured_snr_values}")
    return disturbance_models


class OxfordMARunner(Runner):
    def __init__(self, params: RunningSessionParams) -> None:
        self.params = params

    async def run(  # pylint: disable=too-many-locals, too-many-statements
        self,
    ) -> SessionResults:
        if self.params.extra is None:
            raise ValueError("Extra parameters are required for this runner.")
        entropy_gamma_values = self.params.extra["entropy_gamma_values"]
        noise_levels_snr = self.params.extra["noise_levels_snr"]

        model_results: list[ModelResult] = []

        disturbance_models = fetch_models(noise_levels_snr)

        train, val, _ = get_data_multiple_annotators(
            annotation_models=disturbance_models,
            target_shape=self.params.target_img_shape,
            batch_size=self.params.batch_size,
        )

        train_iterable = train.cache()
        val_iterable = val.cache()

        for entropy_gamma in entropy_gamma_values:
            print(f"========== Entropy Gamma: {entropy_gamma} ==========")
            optimizer = Adam()
            model_name = f"UNET-TGCE_SS_gamma_{entropy_gamma}"
            loss_fn = TcgeSs(
                q=0.1,
                config=TcgeConfig(
                    num_annotators=self.params.num_annotators, gamma=entropy_gamma
                ),
            )
            dice_fn = DiceCoefficient()
            losses = []
            dices = []
            val_losses = []
            val_dices = []

            model = unet_tgce(
                input_shape=self.params.target_img_shape + (3,),
                out_channels=2,
                n_scorers=self.params.num_annotators,
                name=model_name,
            )
            model.compile(loss=loss_fn, metrics=dice_fn, optimizer=optimizer)

            for epoch in range(self.params.n_epochs):
                print(f"Epoch {epoch+1}/{self.params.n_epochs}")
                epoch_loss = 0.0
                epoch_dice = 0.0
                epoch_val_loss = 0.0
                epoch_val_dice = 0.0
                epoch_losses = []
                epoch_dices = []
                for x_batch, y_batch in train_iterable:
                    with tf.GradientTape() as tape:
                        predictions = model(x_batch, training=True)
                        loss = loss_fn.call(y_batch, predictions)

                    gradients = tape.gradient(loss, model.trainable_variables)
                    optimizer.apply_gradients(
                        zip(gradients, model.trainable_variables)  # type:ignore
                    )

                    epoch_loss += float(loss.numpy())
                    epoch_losses.append(loss.numpy())
                    y_pred_ = predictions[..., :2]
                    y_pred_ = y_pred_[..., tf.newaxis]
                    y_pred_ = tf.repeat(
                        y_pred_, repeats=[self.params.num_annotators], axis=-1
                    )
                    dice = tf.reduce_mean(dice_fn.call(y_batch, y_pred_))
                    epoch_dice += float(dice.numpy())
                    epoch_dices.append(dice.numpy())
                losses.append(np.array(epoch_losses).mean())
                dices.append(np.array(epoch_dices).mean())

                epoch_val_losses = []
                epoch_val_dices = []

                for x_batch_val, y_batch_val in val_iterable:
                    val_predictions = model(x_batch_val, training=False)
                    val_loss = loss_fn.call(y_batch_val, val_predictions)
                    val_total_loss = val_loss

                    epoch_val_loss += float(val_total_loss.numpy())
                    y_pred_val = val_predictions[..., :2]
                    y_pred_val = y_pred_val[..., tf.newaxis]
                    y_pred_val = tf.repeat(
                        y_pred_val, repeats=[self.params.num_annotators], axis=-1
                    )
                    val_dice = tf.reduce_mean(dice_fn.call(y_batch_val, y_pred_val))
                    epoch_val_dice += float(val_dice.numpy())

                    epoch_val_losses.append(val_total_loss.numpy())
                    epoch_val_dices.append(val_dice.numpy())

                val_losses.append(np.array(epoch_val_losses).mean())
                val_dices.append(np.array(epoch_val_dices).mean())

                print(
                    f"Training Loss: {np.array(epoch_losses).mean():.3f}, "
                    f"Training Dice: {np.array(epoch_dices).mean():.3f}"
                )
                y_batch = None
                x_batch = None
                print(
                    f"Validation Loss: {np.array(epoch_val_losses).mean():.3f}, "
                    f"Validation Dice: {np.array(epoch_val_dices).mean():.3f}"
                )
                if epoch % self.params.plotting_frequency == 0:
                    epoch_progress_plotter(
                        x_batch,
                        y_batch,
                        predictions,
                        np.random.randint(0, len(x_batch) - 1),
                        noise_levels_snr,
                        self.params.num_annotators,
                    )

            model_result = ModelResult(losses, dices, val_losses, val_dices)
            file_name = f"{model_name}.json"
            model_result.save_as_json(file_name)
            model_results.append(model_result)
            trained_model_path = f"{model_name}.keras"
            model.save(trained_model_path)
            plot_losses_and_metrics(losses, dices, val_losses, val_dices, model_name)

        return SessionResults(models=[], train_metadata={})  # type:ignore

    async def stop(self) -> SessionPartialResults:
        print("Stopping!")
        return SessionPartialResults(train_metadata={})



from keras.models import Model
from tensorflow import GradientTape


class ModelMultipleAnnotators(
    Model
):  # pylint: disable=abstract-method, too-few-public-methods
    def train_step(self, data):
        x, y = data

        with GradientTape() as tape:
            y_pred = self(x, training=True)
            loss, _, _ = self.compute_loss(y=y, y_pred=y_pred)

        gradients = tape.gradient(loss, self.trainable_vars)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_vars))
        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(y, y_pred)
        return {m.name: m.result() for m in self.metrics}



import keras.backend as K
from keras.layers import Layer
from tensorflow import Tensor


class SparseSoftmax(Layer):
    """Custom layer implementing the sparse softmax activation function."""

    def _init_(self, name="SparseSoftmax", **kwargs):
        super()._init_(name=name, **kwargs)

    def call(self, inputs: Tensor) -> Tensor:  # pylint: disable=arguments-differ
        e_x = K.exp(inputs - K.max(inputs, axis=-1, keepdims=True))
        sum_e_x = K.sum(e_x, axis=-1, keepdims=True)
        output = e_x / (sum_e_x + K.epsilon())
        return output



from seg_tgce.data.crowd_seg import get_all_data


def main() -> None:
    train, val, test = get_all_data(batch_size=8, with_sparse_data=True)
    val.visualize_sample(batch_index=138, sample_indexes=[2, 3, 4, 5])
    print(f"Train: {len(train)} batches, {len(train) * train.batch_size} samples")
    print(f"Val: {len(val)} batches, {len(val) * val.batch_size} samples")
    print(f"Test: {len(test)} batches, {len(test) * test.batch_size} samples")

    for i in range(1):
        img, mask = train[i]
        print(f"Batch {i}: {img.shape}, {mask.shape}")


main()



import os
from typing import Tuple

import matplotlib.pyplot as plt
from cv2 import imread  # pylint: disable=no-name-in-module


class BaseDirectoryNotFoundError(Exception):
    pass


def visualize_data(  # pylint: disable=too-many-locals
    x_ini_values: Tuple[int, ...],
    y_ini_values: Tuple[int, ...],
    labelers: Tuple[str, str],
    base_path: str,
    save_path: str,
) -> None:
    """
    Simple routine for visualizing some patches and masks
    """

    fig, axes = plt.subplots(len(x_ini_values), 3)

    for i, (x_ini, y_ini) in enumerate(zip(x_ini_values, y_ini_values)):
        if not os.path.exists(base_path):
            raise BaseDirectoryNotFoundError(
                f"Could not find base directory: {base_path}"
            )

        img_path = (
            f"{base_path}/patches/Train/core_A0AL_AN_x_ini_{x_ini}_y_ini_{y_ini}.png"
        )
        non_expert_mask_path = (
            f"{base_path}/masks/Train/{labelers[0]}/"
            f"core_A0AL_AN_x_ini_{x_ini}_y_ini_{y_ini}.png"
        )
        expert_mask_path = (
            f"{base_path}/masks/Train/{labelers[1]}/"
            f"core_A0AL_AN_x_ini_{x_ini}_y_ini_{y_ini}.png"
        )

        img = imread(img_path)
        non_expert_mask = imread(non_expert_mask_path, -1)
        expert_mask = imread(expert_mask_path, -1)

        axes[i, 0].imshow(img)
        axes[i, 0].axis("off")
        axes[i, 1].imshow(non_expert_mask, cmap="Pastel1")
        axes[i, 1].axis("off")
        axes[i, 2].imshow(expert_mask, cmap="Pastel1")
        axes[i, 2].axis("off")

        if i == 0:
            axes[i, 0].set_title("Histology patch")
            axes[i, 1].set_title("Non expert label mask")
            axes[i, 2].set_title("Expert label mask")
    fig.savefig(save_path)
    plt.show()


if __name__ == "__main__":
    x_ini_values = 1074, 1432, 2148
    y_ini_values = 1074, 1432, 2148
    labelers = "NP1", "expert"
    BASE_PATH = "../../../datasets/Histology Data"
    SAVE_PATH = "../docs/source/resources/crowd-seg-example-instances.png"
    visualize_data(
        x_ini_values=x_ini_values,
        y_ini_values=y_ini_values,
        labelers=labelers,
        base_path=BASE_PATH,
        save_path=SAVE_PATH,
    )



from dataclasses import dataclass
from typing import Any

import keras.backend as K
import tensorflow as tf
from keras.losses import Loss
from tensorflow import cast
from tensorflow import float32 as tf_float32

TARGET_DATA_TYPE = tf_float32


@dataclass
class TcgeConfig:
    """
    TCGE configuration parameters.
    """

    num_annotators: int = 5
    num_classes: int = 2
    gamma: float = 0.1


def binary_entropy(target: tf.Tensor, pred: tf.Tensor) -> tf.Tensor:
    """
    Adds binary entropy to the loss.
    """
    pred_probs = tf.sigmoid(pred)
    hadamard_product = target * pred_probs
    epsilon = 1e-12
    entropy = -hadamard_product * K.log(hadamard_product + epsilon) - (
        1 - hadamard_product
    ) * K.log(1 - hadamard_product + epsilon)
    return K.mean(entropy)


class TcgeSs(Loss):  # type: ignore
    """
    Truncated generalized cross entropy
    for semantic segmentation loss.
    """

    def __init__(
        self,
        config: TcgeConfig,
        q: float = 0.1,
        name: str = "TGCE_SS",
        smooth: float = 1e-5,
    ) -> None:
        self.q = q
        self.num_annotators = config.num_annotators
        self.num_classes = config.num_classes
        self.smooth = smooth
        self.gamma = config.gamma
        super().__init__(name=name)

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        y_true = cast(y_true, TARGET_DATA_TYPE)
        y_pred = cast(y_pred, TARGET_DATA_TYPE)

        y_pred = y_pred[..., : self.num_classes + self.num_annotators]  # type:ignore
        y_true = tf.reshape(
            y_true, (y_true.shape[:-1]) + (self.num_classes, self.num_annotators)
        )
        lambda_r = y_pred[..., self.num_classes :]  # type:ignore
        y_pred_ = y_pred[..., : self.num_classes]
        n_samples, width, height, _ = y_pred_.shape
        y_pred_ = y_pred_[..., tf.newaxis]  # type:ignore
        y_pred_ = tf.repeat(y_pred_, repeats=[self.num_annotators], axis=-1)

        epsilon = 1e-8
        y_pred_ = tf.clip_by_value(y_pred_, epsilon, 1.0 - epsilon)

        term_r = tf.math.reduce_mean(
            tf.math.multiply(
                y_true,
                (
                    tf.ones(
                        [
                            n_samples,
                            width,
                            height,
                            self.num_classes,
                            self.num_annotators,
                        ]
                    )
                    - tf.pow(y_pred_, self.q)
                )
                / (self.q + epsilon + self.smooth),
            ),
            axis=-2,
        )

        term_c = tf.math.multiply(
            tf.ones([n_samples, width, height, self.num_annotators]) - lambda_r,
            (
                tf.ones([n_samples, width, height, self.num_annotators])
                - tf.pow(
                    (1 / self.num_classes + self.smooth)
                    * tf.ones([n_samples, width, height, self.num_annotators]),
                    self.q,
                )
            )
            / (self.q + epsilon + self.smooth),
        )

        loss = tf.math.reduce_mean(tf.math.multiply(lambda_r, term_r) + term_c)
        loss = tf.where(tf.math.is_nan(loss), tf.constant(1e-8), loss)

        return loss

    def get_config(
        self,
    ) -> Any:
        """
        Retrieves loss configuration.
        """
        base_config = super().get_config()
        return {**base_config, "q": self.q}


class TcgeSsSparse(TcgeSs):
    """
    Truncated generalized cross entropy
    for semantic segmentation loss.
    This is a much more sparse version which completes missing dimensions
    for facilitating its use in the Bayesian U-Net.
    """

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        y_true = cast(y_true, TARGET_DATA_TYPE)
        y_pred = cast(y_pred, TARGET_DATA_TYPE)

        y_pred = y_pred[..., : self.num_classes + self.num_annotators]  # type:ignore

        y_true_shape = tf.shape(y_true)

        new_shape = tf.concat(
            [y_true_shape[:-1], [self.num_classes, self.num_annotators]], axis=0
        )
        y_true = tf.reshape(y_true, new_shape)

        lambda_r = y_pred[..., self.num_classes :]  # type:ignore
        y_pred_ = y_pred[..., : self.num_classes]

        n_samples = tf.shape(y_pred_)[0]
        width = tf.shape(y_pred_)[1]
        height = tf.shape(y_pred_)[2]

        y_pred_ = y_pred_[..., tf.newaxis]  # type:ignore
        y_pred_ = tf.repeat(y_pred_, repeats=[self.num_annotators], axis=-1)

        epsilon = 1e-8
        y_pred_ = tf.clip_by_value(y_pred_, epsilon, 1.0 - epsilon)

        term_r = tf.math.reduce_mean(
            tf.math.multiply(
                y_true,
                (
                    tf.ones(
                        [
                            n_samples,
                            width,
                            height,
                            self.num_classes,
                            self.num_annotators,
                        ]
                    )
                    - tf.pow(y_pred_, self.q)
                )
                / (self.q + epsilon + self.smooth),
            ),
            axis=-2,
        )

        term_c = tf.math.multiply(
            tf.ones([n_samples, width, height, self.num_annotators]) - lambda_r,
            (
                tf.ones([n_samples, width, height, self.num_annotators])
                - tf.pow(
                    (1 / self.num_classes + self.smooth)
                    * tf.ones([n_samples, width, height, self.num_annotators]),
                    self.q,
                )
            )
            / (self.q + epsilon + self.smooth),
        )

        loss = tf.math.reduce_mean(tf.math.multiply(lambda_r, term_r) + term_c)
        loss = tf.where(tf.math.is_nan(loss), tf.constant(1e-8), loss)

        return loss



from typing import Tuple

import tensorflow as tf
from keras.models import Model
from tensorflow import Tensor


def disturb_mask(
    model: Model,
    image: Tensor,
    model_shape: Tuple[int, int],
    target_shape: Tuple[int, int],
) -> Tensor:
    return tf.image.resize(model(tf.image.resize(image, model_shape)), target_shape)


def mix_channels(mask: Tensor) -> Tensor:
    return tf.stack([mask, 1 - mask], axis=-2)


def add_noisy_annotators(
    img: Tensor,
    models: list[Tensor],
    model_shape: Tuple[int, int],
    target_shape: Tuple[int, int],
) -> Tensor:
    return tf.transpose(
        [
            disturb_mask(model, img, model_shape=model_shape, target_shape=target_shape)
            for model in models
        ],
        [2, 3, 1, 4, 0],
    )


def map_dataset_multiple_annotators(
    dataset: Tensor,
    target_shape: tuple[int, int],
    model_shape: tuple[int, int],
    batch_size: int,
    disturbance_models: list[Model],
) -> Tensor:
    dataset_ = dataset.map(
        lambda img, mask, label, id_img: (img, mask),
        num_parallel_calls=tf.data.AUTOTUNE,
    )

    dataset_ = dataset_.map(
        lambda img, mask: (
            tf.image.resize(img, target_shape),
            tf.image.resize(mask, target_shape),
        ),
        num_parallel_calls=tf.data.AUTOTUNE,
    )

    dataset_ = dataset_.map(
        lambda img, mask: (
            img,
            add_noisy_annotators(
                tf.expand_dims(img, 0),
                disturbance_models,
                model_shape=model_shape,
                target_shape=target_shape,
            ),
        ),
        num_parallel_calls=tf.data.AUTOTUNE,
    )

    dataset_ = dataset_.map(
        lambda img, mask: (
            img,
            tf.reshape(mask, (mask.shape[0], mask.shape[1], 1, mask.shape[-1])),
        ),
        num_parallel_calls=tf.data.AUTOTUNE,
    )

    dataset_ = dataset_.map(
        lambda img, mask: (img, mix_channels(mask)),
        num_parallel_calls=tf.data.AUTOTUNE,
    )

    dataset_ = dataset_.map(
        lambda img, mask: (img, tf.squeeze(mask, axis=2)),
        num_parallel_calls=tf.data.AUTOTUNE,
    )

    dataset_ = dataset_.batch(batch_size)
    return dataset_



import tensorflow as tf
from keras.models import Model

from seg_tgce.data.utils import map_dataset_multiple_annotators

from .oxford_iiit_pet import OxfordIiitPet

MODEL_ORIGINAL_SHAPE = (256, 256)


def get_data_multiple_annotators(
    annotation_models: list[Model],
    target_shape: tuple[int, int] = (256, 256),
    batch_size: int = 32,
) -> tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    dataset = OxfordIiitPet()
    train_dataset, val_dataset, test_dataset = dataset()
    train_data, val_data, test_data = (
        map_dataset_multiple_annotators(
            dataset=data,
            target_shape=target_shape,
            model_shape=MODEL_ORIGINAL_SHAPE,
            batch_size=batch_size,
            disturbance_models=annotation_models,
        )
        for data in (train_dataset, val_dataset, test_dataset)
    )
    return train_data, val_data, test_data



from enum import Enum
import logging
import os
from typing import List, Optional, Tuple, TypedDict

import numpy as np
from keras.preprocessing.image import img_to_array, load_img
from keras.utils import Sequence
from matplotlib import pyplot as plt
from tensorflow import Tensor, transpose, reshape
from tensorflow import argmax as tf_argmax

from .__retrieve import fetch_data, get_masks_dir, get_patches_dir
from .stage import Stage

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.WARNING)

CLASSES_DEFINITION = {
    0: "Ignore",
    1: "Other",
    2: "Tumor",
    3: "Stroma",
    4: "Benign Inflammation",
    5: "Necrosis",
}
DEFAULT_IMG_SIZE = (512, 512)


class ScorerNotFoundError(Exception):
    pass


class InvalidClassLabelError(Exception):
    pass


class CustomPath(TypedDict):
    """Custom path for image and mask directories."""

    image_dir: str
    mask_dir: str


class DataSchema(Enum):
    """Data schema for the dataset.
    MA_RAW: Raw data for multiple annotators.
    MA_SPARSE: Processed data for multiple annotators. Sparse for fulfilling the
    required dimensions for consistency with the model.
    """

    MA_RAW = 0
    MA_SPARSE = 1


def get_image_filenames(image_dir: str) -> List[str]:
    return sorted(
        [filename for filename in os.listdir(image_dir) if filename.endswith(".png")]
    )


class ImageDataGenerator(Sequence):  # pylint: disable=too-many-instance-attributes
    """
    Data generator for crowd segmentation data.
    Delivered data is in the form of images, masks and scorers labels.
    Shapes are as follows:
    - images: (batch_size, image_size[0], image_size[1], 3)
    - masks: (batch_size, image_size[0], image_size[1], n_classes, n_scorers)"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        image_size: Tuple[int, int] = DEFAULT_IMG_SIZE,
        batch_size: int = 32,
        shuffle: bool = False,
        stage: Stage = Stage.TRAIN,
        paths: Optional[CustomPath] = None,
        schema: DataSchema = DataSchema.MA_RAW,
    ) -> None:
        if paths is not None:
            image_dir = paths["image_dir"]
            mask_dir = paths["mask_dir"]
        else:
            fetch_data()
            image_dir = get_patches_dir(stage)
            mask_dir = get_masks_dir(stage)
        self.image_dir = image_dir
        self.mask_dir = mask_dir
        self.image_size = image_size
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.image_filenames = get_image_filenames(image_dir)
        self.scorers_tags = sorted(os.listdir(mask_dir))
        self.on_epoch_end()
        self.schema = schema

    @property
    def classes_definition(self) -> dict[int, str]:
        """Returns classes definition."""
        return CLASSES_DEFINITION

    @property
    def n_classes(self) -> int:
        """Returns number of classes."""
        return len(self.classes_definition)

    @property
    def n_scorers(self) -> int:
        """Returns number of scorers."""
        return len(self.scorers_tags)

    def __len__(self) -> int:
        return int(np.ceil(len(self.image_filenames) / self.batch_size))

    def _get_items_raw(self, index: int) -> Tuple[Tensor, Tensor]:
        batch_filenames = self.image_filenames[
            index * self.batch_size : (index + 1) * self.batch_size
        ]
        return self.__data_generation(batch_filenames)

    def _get_items_sparse(self, index: int) -> Tuple[Tensor, Tensor]:
        batch_filenames = self.image_filenames[
            index * self.batch_size : (index + 1) * self.batch_size
        ]
        images, masks = self.__data_generation(batch_filenames)
        return images, reshape(masks, (self.batch_size, *self.image_size, -1))

    def __getitem__(self, index: int) -> Tuple[Tensor, Tensor]:
        match self.schema:
            case DataSchema.MA_RAW:
                return self._get_items_raw(index)
            case DataSchema.MA_SPARSE:
                return self._get_items_sparse(index)

    def on_epoch_end(self) -> None:
        if self.shuffle:
            np.random.shuffle(self.image_filenames)

    def visualize_sample(
        self,
        batch_index: int = 0,
        sample_indexes: Optional[List[int]] = None,
        scorers: Optional[List[str]] = None,
    ) -> plt.Figure:
        """
        Visualizes a sample from the dataset."""
        if scorers is None:
            scorers = self.scorers_tags
        images, masks = self._get_items_raw(batch_index)
        if sample_indexes is None:
            sample_indexes = [0, 1, 2, 3]

        fig, axes = plt.subplots(len(sample_indexes), len(scorers) + 1)
        for ax in axes.flatten():
            ax.axis("off")

        for i, sample_index in enumerate(sample_indexes):
            axes[i, 0].imshow(images[sample_index].astype(int))
            axes[i, 0].set_title(
                self.image_filenames[batch_index * self.batch_size + sample_index]
            )
            for j, scorer in enumerate(scorers):
                if scorer not in self.scorers_tags:
                    raise ScorerNotFoundError(
                        f"Scorer {scorer} not found in the dataset scorers "
                        f"({self.scorers_tags})."
                    )
                scorer_index = self.scorers_tags.index(scorer)
                axes[i, j + 1].imshow(
                    tf_argmax(masks[sample_index, :, :, :, scorer_index], axis=2),
                    cmap="viridis",
                )
                axes[i, j + 1].set_title(f"Label for {scorer}")

        plt.show()
        return fig

    def __data_generation(self, batch_filenames: List[str]) -> Tuple[Tensor, Tensor]:
        images = np.empty((self.batch_size, *self.image_size, 3))
        masks = np.empty(
            (
                self.batch_size,
                self.n_scorers,
                self.n_classes,
                *self.image_size,
            )
        )

        for batch, filename in enumerate(batch_filenames):
            img_path = os.path.join(self.image_dir, filename)
            for scorer, scorer_dir in enumerate(self.scorers_tags):
                scorer_mask_dir = os.path.join(self.mask_dir, scorer_dir)
                mask_path = os.path.join(scorer_mask_dir, filename)
                if os.path.exists(mask_path):
                    mask_raw = load_img(
                        mask_path,
                        color_mode="grayscale",
                        target_size=self.image_size,
                    )
                    mask = img_to_array(mask_raw)
                    if not np.all(
                        np.isin(np.unique(mask), list(self.classes_definition))
                    ):
                        raise InvalidClassLabelError(
                            f"Mask {mask_path} contains invalid values. "
                            f"Expected values: {list(self.classes_definition)}."
                            f"Values found: {np.unique(mask)}"
                        )
                    for class_num in self.classes_definition:
                        masks[batch][scorer][class_num] = np.where(
                            mask == class_num, 1, 0
                        ).reshape(*self.image_size)
                else:
                    LOGGER.debug(
                        "Mask not found for scorer %s and image %s",
                        scorer_dir,
                        filename,
                    )
                    masks[batch, scorer, 0] = np.ones(self.image_size)
                    masks[batch, scorer, 1:] = np.zeros(
                        (self.n_classes - 1, *self.image_size)
                    )

            image = load_img(img_path, target_size=self.image_size)
            image = img_to_array(image)

            images[batch] = image

        return images, transpose(masks, perm=[0, 3, 4, 2, 1])



# Inspired from https://github.com/UN-GCPDS/python-gcpds.image_segmentation
# Original license: BSD-2-Clause

from functools import cache
from typing import List, Tuple

import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow import Tensor


class OxfordIiitPet:
    def __init__(
        self,
        split: Tuple[float, float, float] = (70.0, 15.0, 15.0),
        one_hot: bool = True,
    ):
        self.one_hot = one_hot
        self.split = OxfordIiitPet._get_splits(split)
        dataset, info = tfds.load(
            "oxford_iiit_pet:3.*.*", with_info=True, split=self.split
        )
        self.info = info
        train, val, test = dataset
        self.classes = 3
        train = train.map(self._keep_interface)
        val = val.map(self._keep_interface)
        test = test.map(self._keep_interface)

        self.mapped_dataset = train, val, test
        self.labels_info = {0: "cat", 1: "dog"}

    @cache  # pylint: disable=method-cache-max-size-none
    def load_instance_by_id(  # type: ignore
        self,
        id_img: str,
    ) -> Tuple[Tensor, Tensor, Tensor, str]:
        for dataset in self.mapped_dataset:
            dataset = dataset.filter(
                lambda img, mask, label, id_image: id_image == id_img
            )
            for x in dataset:
                return x

    @staticmethod
    def _get_splits(splits: Tuple[float, float, float]) -> List[str]:
        percentage_sum = 0.0
        splits_ = []
        for percentage in splits:
            percentage_sum += percentage
            splits_.append(f"train[{percentage_sum-percentage}%:{percentage_sum}%]")
        return splits_

    def to_one_hot(self, mask: Tensor) -> Tensor:
        one_hot = tf.one_hot(mask, self.classes)
        return tf.gather(one_hot, 0, axis=2)

    def _keep_interface(self, x: dict) -> Tuple[Tensor, Tensor, Tensor, str]:
        img = tf.cast(x["image"], tf.float32) / 255.0
        mask = x["segmentation_mask"] - 1
        mask = self.to_one_hot(mask) if self.one_hot else mask
        label = x["species"]
        id_image = x["file_name"]
        return img, mask, label, id_image

    def __call__(
        self,
    ):
        return self.mapped_dataset


if __name__ == "__main__":
    dataset = OxfordIiitPet()
    train_dataset, val_dataset, test_dataset = dataset()
    for img, mask, label, id_img in train_dataset.take(1):
        print(img.shape, mask.shape, label, id_img)
        break
    for img, mask, label, id_img in val_dataset.take(1):
        print(img.shape, mask.shape, label, id_img)
        break
    for img, mask, label, id_img in test_dataset.take(1):
        print(img.shape, mask.shape, label, id_img)
        break



import json
from dataclasses import dataclass


@dataclass
class ModelResult:
    losses: list[float]
    dices: list[float]
    val_losses: list[float]
    val_dices: list[float]

    def save_as_json(self, filename: str) -> None:
        dataclass_dict = self.__dict__
        json_str = json.dumps(dataclass_dict, indent=4)
        with open(filename, "w", encoding="utf-8") as f:
            f.write(json_str)



from functools import partial
from typing import Tuple

from keras.initializers import GlorotUniform
from keras.layers import (
    BatchNormalization,
    Concatenate,
    Conv2D,
    Input,
    Layer,
    MaxPool2D,
    UpSampling2D,
)
from keras.models import Model
from keras.utils import get_custom_objects

from seg_tgce.layers import SparseSoftmax

from .ma_model import ModelMultipleAnnotators

get_custom_objects()["sparse_softmax"] = SparseSoftmax()

DefaultConv2D = partial(Conv2D, kernel_size=3, activation="relu", padding="same")

DefaultPooling = partial(MaxPool2D, pool_size=2)
DilatedConv = partial(
    Conv2D,
    kernel_size=3,
    activation="relu",
    padding="same",
    dilation_rate=10,
    name="DilatedConv",
)


UpSample = partial(UpSampling2D, (2, 2))


def kernel_initializer(seed: float) -> GlorotUniform:
    return GlorotUniform(seed=seed)


def build_encoder(input_layer: Layer) -> Tuple[Layer, Layer, Layer, Layer, Layer]:
    x = BatchNormalization(name="Batch00")(input_layer)
    x = DefaultConv2D(8, kernel_initializer=kernel_initializer(34), name="Conv10")(x)
    x = BatchNormalization(name="Batch10")(x)
    x = level_1 = DefaultConv2D(
        8, kernel_initializer=kernel_initializer(4), name="Conv11"
    )(x)
    x = BatchNormalization(name="Batch11")(x)
    x = DefaultPooling(name="Pool10")(x)  # 128x128 -> 64x64
    x = DefaultConv2D(16, kernel_initializer=kernel_initializer(56), name="Conv20")(x)
    x = BatchNormalization(name="Batch20")(x)
    x = level_2 = DefaultConv2D(
        16, kernel_initializer=kernel_initializer(32), name="Conv21"
    )(x)
    x = BatchNormalization(name="Batch22")(x)
    x = DefaultPooling(name="Pool20")(x)  # 64x64 -> 32x32
    x = DefaultConv2D(32, kernel_initializer=kernel_initializer(87), name="Conv30")(x)
    x = BatchNormalization(name="Batch30")(x)
    x = level_3 = DefaultConv2D(
        32, kernel_initializer=kernel_initializer(30), name="Conv31"
    )(x)
    x = BatchNormalization(name="Batch31")(x)
    x = DefaultPooling(name="Pool30")(x)  # 32x32 -> 16x16
    x = DefaultConv2D(64, kernel_initializer=kernel_initializer(79), name="Conv40")(x)
    x = BatchNormalization(name="Batch40")(x)
    x = level_4 = DefaultConv2D(
        64, kernel_initializer=kernel_initializer(81), name="Conv41"
    )(x)
    x = BatchNormalization(name="Batch41")(x)
    x = DefaultPooling(name="Pool40")(x)  # 16x16 -> 8x8
    return x, level_1, level_2, level_3, level_4


def build_decoder(
    x: Layer, level_1: Layer, level_2: Layer, level_3: Layer, level_4: Layer
) -> Layer:
    x = DefaultConv2D(128, kernel_initializer=kernel_initializer(89), name="Conv50")(x)
    x = BatchNormalization(name="Batch50")(x)
    x = DefaultConv2D(128, kernel_initializer=kernel_initializer(42), name="Conv51")(x)
    x = BatchNormalization(name="Batch51")(x)
    x = UpSample(name="Up60")(x)  # 8x8 -> 16x16
    x = Concatenate(name="Concat60")([level_4, x])
    x = DefaultConv2D(64, kernel_initializer=kernel_initializer(91), name="Conv60")(x)
    x = BatchNormalization(name="Batch60")(x)
    x = DefaultConv2D(64, kernel_initializer=kernel_initializer(47), name="Conv61")(x)
    x = BatchNormalization(name="Batch61")(x)
    x = UpSample(name="Up70")(x)  # 16x16 -> 32x32
    x = Concatenate(name="Concat70")([level_3, x])
    x = DefaultConv2D(32, kernel_initializer=kernel_initializer(21), name="Conv70")(x)
    x = BatchNormalization(name="Batch70")(x)
    x = DefaultConv2D(32, kernel_initializer=kernel_initializer(96), name="Conv71")(x)
    x = BatchNormalization(name="Batch71")(x)
    x = UpSample(name="Up80")(x)  # 32x32 -> 64x64
    x = Concatenate(name="Concat80")([level_2, x])
    x = DefaultConv2D(16, kernel_initializer=kernel_initializer(96), name="Conv80")(x)
    x = BatchNormalization(name="Batch80")(x)
    x = DefaultConv2D(16, kernel_initializer=kernel_initializer(98), name="Conv81")(x)
    x = BatchNormalization(name="Batch81")(x)
    x = UpSample(name="Up90")(x)  # 64x64 -> 128x128
    x = Concatenate(name="Concat90")([level_1, x])
    x = DefaultConv2D(8, kernel_initializer=kernel_initializer(35), name="Conv90")(x)
    x = BatchNormalization(name="Batch90")(x)
    x = DefaultConv2D(8, kernel_initializer=kernel_initializer(7), name="Conv91")(x)
    x = BatchNormalization(name="Batch91")(x)
    return x


def unet_tgce(
    input_shape: Tuple[int, int, int],
    name: str = "UNET",
    out_channels: int = 2,
    n_scorers: int = 5,
    out_act_functions: Tuple[str, str] = ("softmax", "sigmoid"),
) -> ModelMultipleAnnotators:
    input_layer = Input(shape=input_shape)
    x, level_1, level_2, level_3, level_4 = build_encoder(input_layer)
    x = build_decoder(x, level_1, level_2, level_3, level_4)
    xy = DefaultConv2D(
        out_channels,
        kernel_size=(1, 1),
        activation=out_act_functions[0],
        kernel_initializer=kernel_initializer(42),
        name="Conv100",
    )(x)
    x_lambda = DefaultConv2D(
        n_scorers,
        kernel_size=(1, 1),
        activation=out_act_functions[1],
        kernel_initializer=kernel_initializer(42),
        name="Conv101-Lambda",
    )(x)
    y = Concatenate()([xy, x_lambda])
    return ModelMultipleAnnotators(input_layer, y, name=name)


def bayesian_unet_tgce(  # pylint: disable=too-many-arguments, too-many-locals
    input_shape: Tuple[int, int, int],
    *,
    name: str = "UNET",
    out_channels: int = 2,
    out_filler_channels: int = 2,
    n_scorers: int = 5,
    out_act_functions: Tuple[str, str] = ("sparse_softmax", "sparse_softmax"),
) -> Model:
    input_layer = Input(shape=input_shape)
    x, level_1, level_2, level_3, level_4 = build_encoder(input_layer)
    x = build_decoder(x, level_1, level_2, level_3, level_4)

    xy = DefaultConv2D(
        out_channels,
        kernel_size=(1, 1),
        activation=out_act_functions[0],
        kernel_initializer=kernel_initializer(42),
        name="Conv100",
    )(x)
    xyy = DefaultConv2D(
        out_filler_channels,
        kernel_size=(1, 1),
        activation=out_act_functions[0],
        kernel_initializer=kernel_initializer(42),
        name="Conv101",
    )(x)
    x_lambda = DilatedConv(
        n_scorers,
        kernel_size=(1, 1),
        activation=out_act_functions[1],
        kernel_initializer=kernel_initializer(42),
        name="Conv101-Lambda",
    )(x)
    y = Concatenate()([xy, x_lambda, xyy])

    return Model(input_layer, y, name=name)



from typing import Optional

import keras.backend as K
from keras.losses import Loss
from keras.saving import get_custom_objects, register_keras_serializable
from tensorflow import Tensor, cast
from tensorflow import float32 as tf_float32
from tensorflow import gather

get_custom_objects().clear()

TARGET_DATA_TYPE = tf_float32


@register_keras_serializable(package="MyLayers")
class DiceCoefficient(Loss):
    """
    Dice coefficient loss for semantic segmentation."""

    def __init__(  # type: ignore
        self,
        smooth: float = 1.0,
        target_class: Optional[int] = None,
        name: str = "DiceCoefficient",
        **kwargs
    ):
        self.smooth = smooth
        self.target_class = target_class
        super().__init__(name=name, **kwargs)

    def call(self, y_true: Tensor, y_pred: Tensor) -> Tensor:
        y_true = cast(y_true, TARGET_DATA_TYPE)
        y_pred = cast(y_pred, TARGET_DATA_TYPE)
        intersection = K.sum(y_true * y_pred, axis=[1, 2])
        union = K.sum(y_true, axis=[1, 2]) + K.sum(y_pred, axis=[1, 2])
        dice_coef = -(2.0 * intersection + self.smooth) / (union + self.smooth)

        if self.target_class is not None:
            dice_coef = gather(dice_coef, self.target_class, axis=1)
        else:
            dice_coef = K.mean(dice_coef, axis=-1)

        return dice_coef

    def get_config(
        self,
    ) -> dict:
        base_config = super().get_config()
        return {**base_config, "smooth": self.smooth, "target_class": self.target_class}



import os
from enum import Enum

import gdown  # type:ignore
import numpy as np
from keras.layers import Conv2D, Layer, UpSampling2D
from keras.models import Model, load_model


def compute_snr(signal: float | np.ndarray, noise_std: float) -> float:
    return float(10 * np.log10(np.mean(signal**2) / noise_std**2))


class SnrType(Enum):
    LOG = 0
    LINEAR = 1


def add_noise_to_layer_weights(
    model: Model,
    layer_num: int,
    noise_snr: float,
    snr_type: SnrType = SnrType.LOG,
    verbose: int = 0,
) -> float:
    layer_weights = model.layers[layer_num].get_weights()

    sig_power = np.mean(layer_weights[0] ** 2)

    if snr_type == SnrType.LOG:
        noise_power = sig_power / (10 ** (noise_snr / 10))
    elif snr_type == SnrType.LINEAR:
        noise_power = sig_power / noise_snr

    noise_std = noise_power ** (1 / 2)

    snr = compute_snr(layer_weights[0], noise_std)

    if verbose > 0:
        print(f"Adding noise for snr: {noise_snr}\n\n")
        print(f"Signal power: {sig_power}")
        print(f"Noise power: {noise_power}\n\n")

    for i in range(layer_weights[0].shape[0]):
        for j in range(layer_weights[0].shape[1]):
            layer_weights[0][i][j] += np.random.randn(128, 128) * noise_std

    model.layers[layer_num].set_weights(layer_weights)
    return snr


def produce_disturbed_models(
    snr_value_list: list[float], base_model_path: str, layer_to_disturb: int
) -> tuple[list[Model], list[float]]:
    snr_measured_values: list[float] = []
    models: list[Model] = []

    for value in snr_value_list:
        model_: Model = load_model(base_model_path, compile=False)
        snr = add_noise_to_layer_weights(model_, layer_to_disturb, value)
        snr_measured_values.append(snr)
        models.append(model_)
    return models, snr_measured_values


def download_base_model() -> str:
    model_file_id = "1x39L3QNDMye1SJhKh1gf4YS-HRFLTs6G"
    gdown.download(id=model_file_id, quiet=False)
    model_extension = "keras"
    paths = []

    for file in os.listdir("."):
        if file.endswith(model_extension):
            paths.append(file)

    return os.path.abspath(paths[0])


def find_last_encoder_conv_layer(model: Model) -> Layer:
    last_conv_encoder_layer = 0
    for i, layer in enumerate(model.layers):
        if isinstance(layer, Conv2D):
            last_conv_encoder_layer = i
        if isinstance(layer, UpSampling2D):
            break
    return last_conv_encoder_layer
